try:
    from sage_lib.master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    from sage_lib.master.AtomicProperties import AtomicProperties
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing AtomicProperties: {str(e)}\n")
    del sys

try:
    from sage_lib.input.KPointsManager import KPointsManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing KPointsManager: {str(e)}\n")
    del sys

try:
    from sage_lib.input.input_handling_tools.InputFile import InputFile
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing InputFile: {str(e)}\n")
    del sys

try: 
    from sage_lib.input.PotentialManager import PotentialManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing PotentialManager: {str(e)}\n")
    del sys

try:
    from sage_lib.input.structure_handling_tools.AtomPosition import AtomPosition
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing AtomPosition: {str(e)}\n")
    del sys

try:
    import re
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing re: {str(e)}\n")
    del sys

try:
    import copy
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing copy: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class OutFileManager(FileManager, AtomicProperties):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        """
        Initialize OutFileManager class.
        :param file_location: Location of the file to be read.
        :param name: Name identifier for the file.
        :param kwargs: Additional keyword arguments.
        """
        FileManager.__init__(self, name=name, file_location=file_location)
        AtomicProperties.__init__(self)

        self._comment = None
        self._InputFileManager = InputFile(self.file_location)
        self._KPointsManager = KPointsManager(self.file_location)
        self._AtomPositionManager = [ ]
        self._PotentialManager = PotentialManager(self.file_location)
        
        self._parameters = {}
        self._parameters_data = ['SYSTEM', 'POSCAR', 'Startparameter for this run:', 'NWRITE', 'PREC', 'ISTART', 
                                'ICHARG', 'ISPIN', 'LNONCOLLINEAR', 'LSORBIT', 'INIWAV', 'LASPH', 'METAGGA', 
                                'Electronic Relaxation 1', 'ENCUT', 'ENINI', 'ENAUG', 'NELM', 'EDIFF', 'LREAL', 
                                'NLSPLINE', 'LCOMPAT', 'GGA_COMPAT', 'LMAXPAW', 'LMAXMIX', 'VOSKOWN', 'ROPT', 
                                'ROPT', 'Ionic relaxation', 'EDIFFG', 'NSW', 'NBLOCK', 'IBRION', 'NFREE', 'ISIF',
                                'IWAVPR', 'ISYM', 'LCORR', 'POTIM', 'TEIN', 'TEBEG', 'SMASS', 'estimated Nose-frequenzy (Omega)', 
                                'SCALEE', 'NPACO', 'PSTRESS', 'Mass of Ions in am', 'POMASS', 'Ionic Valenz', 
                                'ZVAL', 'Atomic Wigner-Seitz radii', 'RWIGS', 'virtual crystal weights', 'VCA', 
                                'NELECT', 'NUPDOWN', 'DOS related values:', 'EMIN', 'EFERMI', 'ISMEAR', 
                                'Electronic relaxation 2 (details)', 'IALGO', 'LDIAG', 'LSUBROT', 'TURBO', 
                                'IRESTART', 'NREBOOT', 'NMIN', 'EREF', 'IMIX', 'AMIX', 'AMIX_MAG', 'AMIN', 'WC', 
                                'Intra band minimization:', 'WEIMIN', 'EBREAK', 'DEPER', 'TIME', 'volume/ion in A,a.u.', 
                                'Fermi-wavevector in a.u.,A,eV,Ry', 'Thomas-Fermi vector in A', 'Write flags', 
                                'LWAVE', 'LCHARG', 'LVTOT', 'LVHAR', 'LELF', 'LORBIT', 'Dipole corrections', 
                                'LMONO', 'LDIPOL', 'IDIPOL', 'EPSILON', 'Exchange correlation treatment:', 'GGA', 
                                'LEXCH', 'VOSKOWN', 'LHFCALC', 'LHFONE', 'AEXX', 'Linear response parameters', 
                                'LEPSILON', 'LRPA', 'LNABLA', 'LVEL', 'LINTERFAST', 'KINTER', 'CSHIFT', 'OMEGAMAX', 
                                'DEG_THRESHOLD', 'RTIME', 'Orbital magnetization related:', 'ORBITALMAG', 'LCHIMAG', 'TITEL', 
                                'DQ','type', 'uniqueAtomLabels', 'NIONS', 'atomCountByType']

    def _extract_variables(self, value):
        tokens = re.split(r'[ \t]+', value.strip())
        processed = []
        
        for i, t in enumerate(tokens):
            token_value = 'True' if t == 'T' else 'False' if t == 'F' else str(t) if self.is_number(t) else t
            if i > 0 and not self.is_number(t): break
            processed.append(token_value)

        return ' '.join(processed) if len(processed) > 1 else processed[0]


    def readOUTCAR(self, file_location:str=None):
        
        def _extract_parameter(APM_attribute, initial_j, columns_slice=slice(1, None)):
            j = initial_j
            setattr(APM, APM_attribute, [])
            while True:
                if not lines[i + j].strip():break
                try:
                    getattr(APM, APM_attribute).append( list(map(float, lines[i + j].split()))[columns_slice] )
                    j += 1
                except:
                    break
            setattr(APM, APM_attribute, np.array(getattr(APM, APM_attribute)))

            return APM

        file_location = file_location if type(file_location) == str else self.file_location
        lines =list(self.read_file(file_location,strip=False))
        
        # Make frequently used methods local
        local_strip = str.strip
        local_split = str.split

        read_parameters = True
        APM_holder = []
        APM = None 
        uniqueAtomLabels = []

        # Precompile regular expressions for faster matching
        param_re = re.compile(r"(\w+)\s*=\s*([^=]+)(?:\s+|$)")
        keyword_re = re.compile(r'E-fermi|POTCAR|total charge|magnetization|TOTAL-FORCE|energy  without entropy=|Edisp|Ionic step|direct lattice vectors')
        keyword_ion = re.compile(r'E-Ionic step|Iteration')

        for i, line in enumerate(lines):
            stripped_line = line.strip()
            line_vec = [x for x in line.strip().split(' ') if x]
            
            if read_parameters:
                if keyword_ion.search(line):
                    read_parameters = False

                    self.parameters['uniqueAtomLabels'] = uniqueAtomLabels[:len(uniqueAtomLabels)//2]
                    self.parameters['atomCountByType'] = self.parameters['type']
                    APM = AtomPosition()
                    APM._atomCount = self.parameters['NIONS']
                    APM._uniqueAtomLabels = self.parameters.get('uniqueAtomLabels')
                    APM._atomCountByType = [ int(n) for n in self.parameters.get('atomCountByType').split(' ') ]

                elif 'POTCAR:' in line:
                    uniqueAtomLabels.append( list(set(re.split(r'[ _]', line)).intersection(self.atomic_id))[0] )
                elif 'NIONS' in line:
                    self.parameters['NIONS'] = int(line_vec[11])

                else:
                    for key, value in re.findall(param_re, line.strip()):
                        self._update_parameters(key, self._extract_variables(value))
  
            elif keyword_re.search(line):  # Usa el método `search` para buscar cualquiera de las palabras clave en la línea
                    
                if 'Ionic step' in line:
                    APM_holder.append(APM)
                    APM = PeriodicSystem()
                    
                    APM._atomCount = self.parameters['NIONS']
                    APM._uniqueAtomLabels = self.parameters.get('uniqueAtomLabels')
                    APM._atomCountByType = [ int(n) for n in self.parameters.get('atomCountByType').split(' ') ]

                elif 'E-fermi' in line:
                    APM._E_fermi = float(line_vec[2])
                elif 'Edisp' in line:
                    APM._Edisp = float(line_vec[-1][:-1])
                elif 'energy  without entropy=' in line:
                    APM._E = float(line_vec[-1])
                elif 'direct lattice vectors' in line: 
                    _extract_parameter('_latticeVectors', 1, slice(0, 3))
                    
                elif 'total charge' in line and not 'charge-density' in line: 
                    _extract_parameter('_charge', 4, slice(1, None))

                elif 'magnetization (x)' in line: 
                    _extract_parameter('_magnetization', 4, slice(1, None))

                elif 'TOTAL-FORCE' in line: 
                    _extract_parameter('_total_force', 2, slice(3, None))
                    _extract_parameter('_atomPositions', 2, slice(0, 3))

                elif '2PiTHz' in line: 
                    _extract_parameter('_IRdisplacement', 2, slice(3, None))

        if APM is not None and not APM in APM_holder: 
            APM_holder.append(APM)

        self._AtomPositionManager = APM_holder

    def _update_parameters(self, var_name, value):
        if var_name in self.parameters_data:
            self.parameters[var_name] = value

        if var_name in self._InputFileManager.parameters_data:
            self._InputFileManager.parameters[var_name] = value

    def _handle_regex_matches(self, line_vec, line):
        if 'E-fermi' in line:
            self.E_fermi = float(line_vec[2])
        elif 'NIONS' in line:
            self.NIONS = int(line_vec[11])
            self._AtomPositionManager[-1]._atomCount = int(line_vec[11])

        elif 'POTCAR' in line: 
            try:    
                if not line.split(':')[1] in self.POTCAR_full:
                    self.POTCAR_full.append(line.split(':')[1])
                if not line.split(':')[1].strip().split(' ')[1] in self.POTCAR:
                    self.POTCAR.append( line.split(':')[1].strip().split(' ')[1] )
            except: pass

        elif 'direct lattice vectors' in line: 
            self._AtomPositionManager[-1]._latticeVectors = np.array((map(float, lines[i + 1 + j].split()[3:])))
            
        elif 'total charge' in line and not 'charge-density' in line: 
            TC = self.NIONS+4
            total_charge = np.zeros((self.NIONS,4))
            TC_counter = 0 

    def export_configXYZ(self, file_location:str=None, save_to_file:str='w', verbose:bool=False):
        file_location  = file_location if file_location else self.file_location+'_config.xyz'

        #f filter_by_forces:
        #for i, APM in enumerate(self.AtomPositionManager):
        #    if np.any(APM.total_force > 10): 
        #        print(file_location, f'iteration {i}')
        #        return False

        with open(file_location, save_to_file) as f:
            for APM in self.AtomPositionManager:
                f.write( APM.export_as_xyz(file_location, save_to_file=False, verbose=False) )

        if verbose:
            print(f"XYZ content has been saved to {file_location}")

'''
print(123)

o = OutFileManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/OUTCAR/OUTCAR')
o.readOUTCAR()
o.export_configXYZ(filter_by_forces=True)

#print( o.InputFileManager.exportAsINCAR() )


print(o.AtomPositionManager)
#print( [ APM.E for APM in o._AtomPositionManager ])
for n in o.AtomPositionManager:
    print(n.total_force[0])

'''



try:
    from sage_lib.master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    from sage_lib.master.AtomicProperties import AtomicProperties
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing AtomicProperties: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

try:
    import json
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing json: {str(e)}\n")
    del sys

class EigenvalueFileManager(FileManager, AtomicProperties):
    def __init__(self, file_location:str=None, name:str=None, cell:np.array=None, fermi:float=None, **kwargs):
        """
        Initialize OutFileManager class.
        :param file_location: Location of the file to be read.
        :param name: Name identifier for the file.
        :param kwargs: Additional keyword arguments.
        """
        FileManager.__init__(self, name=name, file_location=file_location)
        AtomicProperties.__init__(self)
        self._comment = None

        self._bands = None
        self._kpoints = None
        self._k_distance = None

        self._fermi = fermi
        self._cell = cell

    @property
    def cell(self):
        if self._cell is not None:
            return np.array(self._cell, dtype=np.float64)
        else:
            return None

    def read_EIGENVAL(self, file_location:str=None):
        file_location = file_location if type(file_location) == str else self._file_location
        lines = [n for n in self.read_file(file_location) ]
        
        self.bands, self.kpoints=[], []
        var = -1
        for i, n in enumerate(lines):

            vec = [float(m) for m in n.split(' ') if self.is_number(m) ] 
            if i == 5: self.bands = np.zeros((int(vec[1]), int(vec[2]), 4))
            if len(vec) == 4 and i>5: self.kpoints.append(vec); var+=1
            if len(vec) == 5 and i>5: self.bands[var,int(vec[0]-1), :] = vec[1:]

        self.k_distance = np.zeros((len(self.kpoints)))
        var = 0
        for n in range(len(self.k_distance)-1): 
            var += ((self.kpoints[n][0]-self.kpoints[n+1][0])**2+(self.kpoints[n][1]-self.kpoints[n+1][1])**2+(self.kpoints[n][2]-self.kpoints[n+1][2])**2)**0.5
            self.k_distance[n+1] = var

        self.kpoints = np.array(self.kpoints)
        self.bands = np.array(self.bands)
        self.k_distance = np.array(self.k_distance)

        return True

    def _ndarray_2_list(self, array):
        return [list(array.shape), str(array.dtype), list(array.flatten(order='C'))]

    def _ndarray_2_dict(self, array):
        return {'__ndarray__':self._ndarray_2_list(array)}

    def _get_specialpoints(self, kpoints:np.array) -> list:
        """Check if points in a kpoints matrix exist in a lattice points dictionary."""
        found_points = []

        for point in kpoints:
            for label, special_lattice_point in self.special_lattice_points.items():
                # Compare only the first three elements (x, y, z coordinates)
                if self.is_close(point[:3], special_lattice_point[:3]):
                    found_points.append( label )
                    break

        return found_points
    
    def _subtract_fermi(self, fermi:float=None):
        fermi = fermi if fermi is not None else self.fermi 

        self.bands[:,:,:2] -= fermi 
        self.fermi = 0
        
        return True

    def _transform_bands(self, bands:np.array=None):
        bands = bands if bands is not None else self.bands[:,:,:1] if self.bands.shape[2] == 1 else self.bands[:,:,:2] 
        return matrix.reshape(1, *bands.shape) if bands.ndim == 2 else (bands.transpose(2, 0, 1).reshape(2, *bands.shape[:2]) if bands.ndim == 3 and bands.shape[2] == 2 else None)

    def export_as_json(self, file_location:str=None, subtract_fermi:bool=True) -> True:
        file_location = file_location if type(file_location) == str else self._file_location+'data.json'

        if subtract_fermi: self._subtract_fermi()

        SP = self._get_specialpoints(self.kpoints)

        # Crear el formato JSON
        json_data = {
            "path": {
                "kpts": self._ndarray_2_dict(self.kpoints[:,:3]),
                "special_points": {sp:self._ndarray_2_dict(self.special_lattice_points[sp]) for sp in SP},
                "labelseq": ''.join(SP),
                "cell": {"array": self._ndarray_2_dict(self.cell), "__ase_objtype__": "cell"},
                "__ase_objtype__": "bandpath"

                    },
            "energies": self._ndarray_2_dict( self._transform_bands() ), # SPIN x KPOINT x Nband
            "reference": self.fermi,
            "__ase_objtype__": "bandstructure"
        }

        self.save_to_json(json_data, file_location)
        
        return True

    def plot(self, file_location:str=None, subtract_fermi:bool=True, save:bool=False, Emin:float=-5, Emax:float=5) -> bool:
        import matplotlib.pyplot as plt

        file_location = file_location if type(file_location) == str else self._file_location+'img_band.png'

        if subtract_fermi: self._subtract_fermi()

        X_distance = [ np.min([k, 0.1]) for k in np.linalg.norm( self.kpoints[1:,:3] - self.kpoints[:-1,:3], axis=1 ) ] 
        X = [0]
        SP = []
        for k, point in enumerate(self.kpoints):

            for label, special_lattice_point in self.special_lattice_points.items():
                # Compare only the first three elements (x, y, z coordinates)
                if self.is_close(point[:3], special_lattice_point[:3]):
                    SP.append( [X[-1], label] )
                    break

            if k < self.kpoints.shape[0]-1: X.append( X[-1]+X_distance[k] )

        SP = np.array(SP)

        # Añadir líneas punteadas verticales
        for pos in SP[:, 0]:
            plt.axvline(x=float(pos), color='gray', linestyle='--', alpha=0.7 ,linewidth=1)
        plt.axhline(y=0, color='green', linestyle='-', alpha=0.2, linewidth=1)

        plt.ylim(-5, 5)

        Y = self.bands[:,:,:1] if self.bands.shape[2] == 1 else self.bands[:,:,:2] 
        plt.xticks(SP[:, 0].astype(np.float32), SP[:, 1])

        plt.plot(X, Y[:,:,0], color=[0.8,0.1,0.1], alpha=0.6, lw=0.7 )
        if self.bands.shape[2] == 4:
            plt.plot(X, Y[:,:,1], color=[0.1,0.1,0.8], alpha=0.6, lw=0.7 )
 

        # Save the plot to a file
        plt.savefig(f"{file_location}", dpi=350)  # Change the filename and format as needed


        # plt.show()



# Supongamos que tienes una matriz de 123x34. Aquí creamos una matriz de ejemplo.

'''
ei = EigenvalueFileManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/EIGENVAL/test/EIGENVAL', fermi=0)
ei.read_EIGENVAL()
ei.fermi = 0
ei.cell = [[1.0,0,0],[1,0,0],[1,0,0]]
ei.plot()
ei.export_as_json()

asdf
print(ei.bands.shape)

path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/EIGENVAL/bs_wz_ZnO.json'
with open(path, 'r') as file:
    data = json.load(file)
print( data['path'].keys() )

path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/EIGENVAL/EIGENVALdata.json'
with open(path, 'r') as file:
    data = json.load(file)
print( data['path'].keys() )

import matplotlib.pyplot as plt
plt.plot( ei.bands[:,:,1] )
plt.show()

#


        self.n_electrons = None
        self.n_kpoints = None
        self.n_bands = None
        self.bands = None
        self.kpoints = None
        self.k_distance = None
'''




try:
    from sage_lib.partition.partition_builder.BandStructure_builder import BandStructure_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing BandStructure_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.partition_builder.Config_builder import Config_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing Config_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.partition_builder.Crystal_builder import Crystal_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing Crystal_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.partition_builder.CrystalDefect_builder import CrystalDefect_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing CrystalDefect_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.partition_builder.ForceField_builder import ForceField_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing ForceField_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.partition_builder.Molecule_builder import Molecule_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing Molecule_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.partition_builder.PositionEditor_builder import PositionEditor_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing PositionEditor_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.partition_builder.SurfaceStates_builder import SurfaceStates_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing SurfaceStates_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.partition_builder.VacuumStates_builder import VacuumStates_builder
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing VacuumStates_builder: {str(e)}\n")
    del sys

try:
    from sage_lib.partition.PartitionManager import PartitionManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing PartitionManager: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class Partition(BandStructure_builder, Config_builder, Crystal_builder, CrystalDefect_builder, ForceField_builder, 
                 Molecule_builder, PositionEditor_builder, SurfaceStates_builder, VacuumStates_builder):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(name=name, file_location=file_location)

    def generate_variants(self, parameter: str, values:np.array=None, file_location: str = None) -> bool:
        containers = []
        directories = ['' for _ in self.containers]
        parameter = parameter.upper().strip()
        for container_index, container in enumerate(self.containers):
            if parameter.upper() == 'KPOINTS':
                containers += self.handleKPoints(container, values, container_index,  file_location) 
                directories[container_index] = 'KPOINTConvergence'

            elif container.InputFileManager and parameter.upper() in container.InputFileManager.parameters_data:
                containers += self.handleInputFile(container, values, parameter,  container_index, file_location)
                directories[container_index] = f'{parameter}_analysis'

            elif parameter.upper() == 'VACANCY':
                containers += self.handleVacancy(container, values, container_index, file_location)
                directories[container_index] = 'Vacancy'

            elif parameter.upper() == 'BAND_STRUCTURE':
                containers += self.handleBandStruture(container, values, container_index, file_location)
                directories[container_index] = 'band_structure'

            elif parameter.upper() == 'RATTLE':
                containers += self.handleRattle(container, values, container_index, file_location)
                directories[container_index] = 'rattle'

        self.containers = containers
        self.generate_master_script_for_all_containers(directories, file_location if not file_location is None else container.file_location )

'''
path = '/home/akaris/Documents/code/Physics/VASP/v6.2/files/dataset/CoFeNiOOH_jingzhu/bulk_NiFe/rattle'
DP = Partition(path)

DP.readVASPFolder(v=False)
DP.generate_variants('rattle', [ {'N':20, 'std':[0.13]} ] )
'''



try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

try:
    import os
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing os: {str(e)}\n")
    del sys

try:
    from datetime import datetime
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing datetime: {str(e)}\n")
    del sys

try:
    import json
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing datetime: {str(e)}\n")
    del sys

class FileManager:
    def __init__(self, file_location:str=None, name:str=None):
        if name is not None and not isinstance(name, str):
            raise TypeError("(!) name must be a string")
        self._name = name
        self._metaData = {}
        self._comment = ''
        if file_location is not None and not isinstance(file_location, str):
            raise TypeError("(!) file_location must be a string")
        self._file_location = file_location   
        self._original_file_location = None
        
        #self._FM_attrs = set(vars(self).keys())
        
    def __getattr__(self, name):
        # Check if the attribute exists with a leading underscore
        private_name = f"_{name}"
        if private_name in self.__dict__:
            return getattr(self, private_name)
        else:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        # Check if the attribute exists with a leading underscore
        private_name = f"_{name}"
        if private_name in self.__dict__:
            setattr(self, private_name, value)
        else:
            super().__setattr__(name, value)
            
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise TypeError("(!) name must be a string")
        self._name = value

    @name.deleter
    def name(self):
        print("Deleting name")
        del self._name

    @property
    def file_location(self):
        return self._file_location

    @file_location.setter
    def file_location(self, value):
        if not isinstance(value, str):
            raise TypeError("(!) file_location must be a string")
        self._file_location = value
        if self._original_file_location == None: 
            self._original_file_location = self._file_location

    @file_location.deleter
    def file_location(self):
        print("Deleting file_location")
        del self._file_location

    @staticmethod
    def is_number(s):
        try:
            float(s)
            return True
        except ValueError:
            return False

    def isINT(self, num): return self.is_number(num) and abs(num - int(num)) < 0.0001 

    def is_close(self, point_a, point_b, threshold:float=1e-5) -> bool:
        """Check if two points are close to each other within a specified threshold."""
        return np.linalg.norm(point_a - point_b) < threshold

    def file_exists(self, file_path:str) -> bool:
        """Check if a file exists."""
        return os.path.exists(file_path)

    def create_directories_for_path(self, path:str):
        """
        Create any directories needed to ensure that the given path is valid.

        Parameters:
        - path (str): The file or directory path to validate.
        """
        if not os.path.exists(path):
            try:
                os.makedirs(path, exist_ok=True)
            except PermissionError:
                print(f"Permission denied: Could not create directory at {path}")
                # Exit or handle the error appropriately
                exit(1)

    def read_file(self, file_location:str=None, strip=True):
        file_location = file_location if type(file_location) == str else self._file_location

        if file_location is None:
            raise ValueError("(!) File location is not set.")
        
        try:
            with open(file_location, 'r') as file:
                if strip:
                    for line in file:
                        yield line.strip()  # strip() elimina espacios y saltos de línea al principio y al final
                else:
                    for line in file:
                        yield line  # strip() elimina espacios y saltos de línea al principio y al final
        except FileNotFoundError:
            raise FileNotFoundError(f"(!) File not found at {file_location}")
        except IOError:
            raise IOError(f"(!) Error reading file at {file_location}")
        except Exception as e:
            print(f"Error inesperado: {e}")

    def save_to_json(self, data, file_path:str) -> bool:
        """
        Save data to a JSON file.

        Args:
            data (dict): The data to be saved as a Python dictionary.
            file_path (str): The path to the JSON file where the data will be saved.

        Returns:
            bool: True if the data was successfully saved, False otherwise.

        Raises:
            FileNotFoundError: If the directory specified in 'file_path' does not exist.
            IOError: If there was an issue writing the data to the JSON file.

        Example:
            data = {"name": "John", "age": 30, "city": "New York"}
            success = save_to_json(data, "data.json")
            if success:
                print("Data saved successfully.")
            else:
                print("Failed to save data.")
        """
        try:
            # Serialize the data as JSON and write it to the file
            with open(file_path, 'w') as json_file:
                json.dump(data, json_file, indent=4)
            return True
        except FileNotFoundError:
            raise FileNotFoundError("The directory specified in 'file_path' does not exist.")
        except MemoryError:
            print("Memory Error occurred")
        except IOError as e:
            print(f"There was an issue writing the data to the JSON file. I/O error({e.errno}): {e.strerror}")
        except Exception as e:
            print(f"Unexpected error: {e}")

        return True

    def loadStoreManager(self, manager_class, file_name:str, attribute_name: str, read_method:str, v:bool=False):
        """Generic function to read a VASP file."""
        file_path = f"{self.file_location}/{file_name}"
        if not self.file_exists(file_path):
            return
        #try:
        manager = manager_class(file_path)
        getattr(manager, read_method)()  # Call the appropriate read method
        setattr(self, attribute_name, manager)
        self._loaded[file_name] = True
        if v:
            print(f'File readed : {file_name}')
        #except Exception as e:
        #    if v:
        #        print(f"ERROR :: Cannot load {file_name}. Reason: {e}")

    def copy(self, file_location:str=None, original_file_location:str=None, buffer_size=1024*1024) -> bool:
        """
        Universal copy method that chooses an available copy method.
        
        Parameters:
        - dest_path (str): The destination path where the file will be copied.
        - buffer_size (int): The buffer size for manual copy if needed.
        
        Returns:
        - bool: True if the file was successfully copied, otherwise False.
        """
        file_location  = file_location if file_location else self.file_location
        original_file_location  = original_file_location if original_file_location else self.original_file_location

        if original_file_location is None:
            print("Original file path is not set.")
            return False
        try:
            os.system(f"cp {original_file_location} {file_location}")
            return True
        except Exception as e:
            if verbosity: print(f"An error occurred while copying the file: {e}")
            return False 

    def copy_file_manual(self, file_location:str=None, original_file_location:str=None, buffer_size:int=1024*1024) -> bool:
        """
        Manually copy a file in a memory-efficient manner.
        
        Parameters:
        - dest_path (str): The destination path where the file will be copied.
        - buffer_size (int): The buffer size for the copy operation.
        
        Returns:
        - bool: True if the file was successfully copied, otherwise False.
        """
        try:
            with open(self.original_file_path, 'rb') as src, open(dest_path, 'wb') as dest:
                chunk = src.read(buffer_size)
                while chunk:
                    dest.write(chunk)
                    chunk = src.read(buffer_size)
            return True
        except Exception as e:
            print(f"An error occurred while copying the file: {e}")
            return False

    def _copy_shared_attributes(self, shared_attributes, source, destination):
        """
        Copies shared attributes from the source object to the destination object.

        :param shared_attributes: List of attribute names to be copied.
        :param source: The source object from which to copy attributes.
        :param destination: The destination object to which attributes are copied.
        """
        for attr in shared_attributes:
            if hasattr(source, attr):
                setattr(destination, attr, getattr(source, attr))




try:
    from sage_lib.partition.PartitionManager import PartitionManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing PartitionManager: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class Config_builder(PartitionManager):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(name=name, file_location=file_location)
        
    def handleKPoints(self, container, values, container_index,  file_location=None):
        sub_directories, containers = [], []

        for v in values:
            container_copy = self.copy_and_update_container(container, f'/KPOINTConvergence/{v[0]}_{v[1]}_{v[2]}', file_location)
            container_copy.KPointsManager.subdivisions = [v[0], v[1], v[2]]
            sub_directories.append(f'{v[0]}_{v[1]}_{v[2]}')
            containers.append(container_copy)

        self.generate_execution_script_for_each_container(sub_directories, container.file_location + '/KPOINTConvergence')
        return containers

    def handleInputFile(self, container, parameter, values, container_index, file_location=None):
        sub_directories, containers = [], []

        for v in values:
            container_copy = self.copy_and_update_container(container, f'/{parameter}_analysis/{v}', file_location)
            container_copy.InputFileManager.parameters[parameter.upper()] = ' '.join(v) if v is list else v 
            sub_directories.append('_'.join(map(str, v)) if isinstance(v, list) else str(v))
            containers.append(container_copy)

        self.generate_execution_script_for_each_container(sub_directories, container.file_location + f'/{parameter}_analysis')
        return containers




try:
    from sage_lib.master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys
    
class DOSManager(FileManager):
    def __init__(self, file_location: str = None, name: str = None, **kwargs):
        """
        Initialize DOSManager with optional file location and name.

        :param file_location: Path to the file containing DOS data.
        :param name: Descriptive name for the DOS data set.
        :param kwargs: Additional keyword arguments for base FileManager.
        """
        super().__init__(name=name, file_location=file_location)
        self._spin_polarized = None  # To be determined based on DOS data
        self._dos_total = None  # Total DOS data
        self._dos_ion = []  # Partial DOS data for each atom
        self._fermi = None
        self._E_fermi = None

    @property
    def E_fermi(self):
        if self._E_fermi is not None:
            return self._E_fermi
        elif self._fermi is not None:
            self._E_fermi = self._fermi
            return self._E_fermi
        else:
            return None

    def _read_total_dos(self, lines):
        """
        Read and parse total DOS data from given lines of a file.

        :param lines: List of string lines containing total DOS data.
        :return: Dictionary with total DOS and integrated DOS data.
        """
        dos_total = {'energies': [], 'dos_up': [], 'dos_down': [],
                     'integrated_dos_up': [], 'integrated_dos_down': []}

        for line in lines:
            values = list(map(float, line.split()))
            if len(values) == 5:
                self.spin_polarized = True 
                dos_total['energies'].append(values[0])
                dos_total['dos_up'].append(values[1])
                dos_total['dos_down'].append(values[2])
                dos_total['integrated_dos_up'].append(values[3])
                dos_total['integrated_dos_down'].append(values[4])
            elif len(values) == 3:
                self.spin_polarized = False
                dos_total['energies'].append(values[0])
                dos_total['dos_up'].append(values[1])  # Assuming 'dos_up' holds total DOS for non-spin-polarized case
                dos_total['integrated_dos_up'].append(values[2])  # Assuming 'integrated_dos_up' holds integrated DOS

        return dos_total

    def _read_ion_dos(self, lines):
        """
        Read and parse partial DOS data for a single ion from given lines of a file.

        :param lines: List of string lines containing partial DOS data for an ion.
        :return: Dictionary with partial DOS data for an ion.
        """
        dos_ion_template = {
            'energies': [],
            's_down': [], 's_up': [],
            'p_x_down': [], 'p_x_up': [],'p_y_down': [], 'p_y_up': [], 'p_z_down': [], 'p_z_up': [],
            'd_xy_down': [], 'd_xy_up': [], 'd_yz_down': [], 'd_yz_up': [], 
            'd_z2r2_down': [], 'd_z2r2_up': [], 'd_xz_down': [], 'd_xz_up': [],
            'd_x2y2_down': [], 'd_x2y2_up': [],
        } if self.spin_polarized else {
            'energies': [],
            's': [],
            'p_x': [], 'p_y': [],'p_z': [],
            'd_xy': [], 'd_yz': [], 'd_z2r2': [], 'd_xz': [], 'd_x2y2': [],
        } 

        dos_ion = {k: [] for k in dos_ion_template}  # Create a dictionary with lists for each orbital
        dos_ion['E_min'], dos_ion['E_max'], dos_ion['NEDOS'], dos_ion['fermi'], _ = list(map(float, lines[0].split()))

        for line in lines[1:]:  # Skip the first line which is header
            values = list(map(float, line.split()))
            if self.spin_polarized:
                for key, value in zip(dos_ion_template, values):
                    dos_ion[key].append(value)
            else:
                for key in dos_ion_template.keys():
                    # If not spin polarized, only every second key should be filled (assuming only 'up' keys are used)
                    dos_ion[key].extend(values[i:i+2] for i in range(0, len(values), 2))

        return dos_ion

    def _read_ions_dos(self, lines):
        """
        Read and parse partial DOS data for all ions from given lines of a file.

        :param lines: List of string lines containing partial DOS data for all ions.
        :return: List of dictionaries with partial DOS data for each ion.
        """
        for n in range(self._atomCount):
            start = (self._NEDOS + 1) * n
            end = start + self._NEDOS + 1
            self.dos_ion.append(self._read_ion_dos(lines[start:end]))

    def _read_header(self, lines):
        """
        Read and parse header information from the file.

        :param lines: First six lines from the file containing header information.
        """
        self._atomCount_spheres, self._atomCount, self._partial_DOS, self._NCDIJ = map(int, lines[0].split())
        self._Volume, self._a, self._b, self._c, self._POTIM = map(float, lines[1].split())
        self._TEBEG = float(lines[2].strip())
        self._name = lines[4].strip()
        self._E_min, self._E_max, self._NEDOS, self._fermi, _ = map(float, lines[5].split())
        self._NEDOS = int(self._NEDOS)

    def read_DOSCAR(self, file_location: str = None) -> bool:
        """
        Load DOS data from a file specified by the file location.

        This method reads the file, extracting header information, total DOS,
        and if available, partial DOS for each ion.

        :param file_location: Path to the file to load. Defaults to initialized file location.
        :return: True if loading succeeds, False otherwise.
        """
        file_location = file_location if type(file_location) == str else self._file_location

        lines = [n for n in self.read_file() ]        
        # Number of Ions (including empty spheres), Number of Ions, 0 (no partial DOS) or 1 (incl. partial DOS), NCDIJ (currently not used)     
        self._read_header(lines[:6])
        self.dos_total = self._read_total_dos(lines[6:6+self._NEDOS])
        self._read_ions_dos(lines[6+self._NEDOS:])

        return True

'''
path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/DOSCAR/surf_NiFeKOH/surf_NiFe_4H_4OH'
DM = DOSManager(path + "/DOSCAR")
DM.load()
print( DM.fermi )
'''


try:
    from sage_lib.master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    from sage_lib.single_run.SingleRun import SingleRun
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing SingleRun: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

try:
    import os 
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing os: {str(e)}\n")
    del sys

try:
    import copy
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing copy: {str(e)}\n")
    del sys

try:
    import re
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing re: {str(e)}\n")
    del sys

class PartitionManager(FileManager): # el nombre no deberia incluir la palabra DFT tieneu qe ser ma general
    """
    A class for partitioning and managing data files for simulations.

    This class extends FileManager and provides functionalities for handling 
    different types of simulation data, applying operations to it, and organizing the results.

    Attributes:
        file_location (str): The file path where the data files are located.
        containers (list): A list of containers to hold various data structures.
    """
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        """
        Initialize the DataPartition instance.

        Parameters:
            file_location (str, optional): Location of the files to be managed.
            name (str, optional): Name of the partition.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(name=name, file_location=file_location)
        self._containers = []

    def add_container(self, container: object):
        """
        Add a new container to the list of containers.

        Parameters:
            container (object): The container object to be added.
        """
        self.containers.append(container)

    def remove_container(self, container: object):
        """
        Remove a container from the list of containers.

        Parameters:
            container (object): The container object to be removed.
        """
        self.containers.remove(container)

    def empty_container(self):
        """
        Empty the list of containers.
        """
        self.containers = []

    def filter_conteiners(self, filter_class:str, value:float) -> bool:
        mask = np.ones_like(containers)
        for container_index, container in enumerate(self.containers):
            if filter_class.lower() == 'fmax':
                if np.any(container.AtomPositionManager.atomPositions > value):
                    mask[container_index] = 0

        self.containers = self.containers[mask]
        return True 

    def readConfigSetup(self, file_location:str=None, source='VASP', v=False):
        file_location = file_location if type(file_location) == str else self.file_location
        
        DFT_SR = self.readVASPFolder(file_location=file_location, add_container=False, v=v)
        DFT_SR.InputFileManager.set_LDAU( DFT_SR.AtomPositionManager.uniqueAtomLabels )

        for c_i, container in enumerate(self.containers):
            container.InputFileManager = DFT_SR.InputFileManager
            container.KPointsManager = DFT_SR.KPointsManager
            container.PotentialManager = DFT_SR.PotentialManager
            container.BashScriptManager = DFT_SR.BashScriptManager
            container.vdw_kernel_Handler = DFT_SR.vdw_kernel_Handler
            container.WaveFileManager = DFT_SR.WaveFileManager
            container.ChargeFileManager = DFT_SR.ChargeFileManager

    def readVASPSubFolder(self, file_location:str=None, v=False):
        file_location = file_location if type(file_location) == str else self.file_location
                
        for root, dirs, files in os.walk(file_location):
            DFT_SR = self.readVASPFolder(file_location=root, add_container=True, v=v)
            if v: print(root, dirs, files)

    def readVASPFolder(self, file_location:str=None, add_container:bool=True, v=False):
        file_location = file_location if type(file_location) == str else self.file_location

        SR = SingleRun(file_location)
        SR.readVASPDirectory()        
        if add_container and SR.AtomPositionManager is not None: 
            self.add_container(container=SR)

        return SR

    def exportVaspPartition(self, file_location:str=None, label:str='fixed'): 
        for c_i, container in enumerate(self.containers):

            if label == 'enumerate':
                container.exportVASP(file_location=file_location+f'/{c_i:03d}')
            if label == 'fixed':
                container.exportVASP(file_location=file_location)

    def read_configXYZ(self, file_location:str=None, verbose:bool=False):
        file_location = file_location if type(file_location) == str else self.file_location

        lines =list(self.read_file(file_location,strip=False))
        container = []

        for i, line in enumerate(lines):
            print(i, line)
            if line.strip().isdigit():
                num_atoms = int(line.strip())
                if num_atoms > 0:
                    DFT_SR = DFTSingleRun(file_location)
                    DFT_SR.AtomPositionManager = PeriodicSystem()
                    DFT_SR.AtomPositionManager.read_configXYZ(lines=lines[i:i+num_atoms+2])

                    container.append(DFT_SR)

        self._containers += container
        return container

    def export_configXYZ(self, file_location:str=None, verbose:bool=False):
        file_location  = file_location if file_location else self.file_location+'_config.xyz'
        with open(file_location, 'w'):pass # Create an empty file
        for container_index, container in enumerate(self.containers):
            if container.OutFileManager is not None:    
                container.OutFileManager.export_configXYZ(file_location=file_location, save_to_file='a', verbose=False)

        if verbose:
            print(f"XYZ content has been saved to {file_location}")

        return True
    
    def _is_redundant(self, containers:list=None, new_container:object=None):
        """
        Checks if the new configuration is redundant within the existing configurations.

        Parameters:
        - configurations (list): List of existing configurations.
        - new_config (object): New configuration to check.

        Returns:
        - bool: True if redundant, False otherwise.
        """
        containers = containers if containers is not None else self.containers
        return any(np.array_equal(conteiner.atomPositions, new_container.atomPositions) for conteiner in containers)

    def summary(self, ) -> str:
        text_str = ''
        text_str += f'{self.file_location}\n'
        text_str += f'> Conteiners : { len(self.containers) }\n'
        return text_str
    
    def copy_and_update_container(self, container, sub_directory: str, file_location=None):
        container_copy = copy.deepcopy(container)
        container_copy.file_location = f'{container.file_location}{sub_directory}' if file_location is None else file_location
        return container_copy

    def generate_execution_script_for_each_container(self, directories: list = None, file_location: str = None):
        self.create_directories_for_path(file_location)
        script_content = self.generate_script_content('RUNscript.sh', directories)
        self.write_script_to_file(script_content, f"{file_location}/execution_script_for_each_container.py")

    def generate_master_script_for_all_containers(self, directories: list = None, file_location: str = None):
        self.create_directories_for_path(file_location)
        script_content = self.generate_script_content('execution_script_for_each_container.py', directories)
        self.write_script_to_file(script_content, f"{file_location}/master_script_for_all_containers.py")

    def generate_script_content(self, script_name: str, directories: list = None) -> str:
        directories_str = "\n".join([f"    '{directory}'," for directory in directories])
        return f'''#!/usr/bin/env python3
import os
import subprocess

original_directory = os.getcwd()

directories = [
{directories_str}
]

for directory in directories:
    os.chdir(directory)
    subprocess.run(['chmod', '+x', '{script_name}'])
    subprocess.run(['sbatch', '{script_name}'])
    os.chdir(original_directory)
'''

    def write_script_to_file(self, script_content: str, file_path: str):
        with open(file_path, "w") as f:
            f.write(script_content)




'''
DP.exportVaspPartition()

print(DP.containers[0].AtomPositionManager.pbc)
DP.generateDFTVariants('band_structure', values=[{'points':20, 'special_points':'GMLCXG'}])
DP.exportVaspPartition()


path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/test/files'
DP = DFTPartition(path)
DP.readVASPSubFolder(v=False)
DP.readConfigSetup('/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/test/config')
DP.generate_execution_script_for_each_container([ f'{n:03d}'for n, c in enumerate(DP.containers) ], '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/test/calcs')

#DP.read_configXYZ()

'''




'''
DP.export_configXYZ()

path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/dataset/CoFeNiOOH_jingzhu/surf_CoFe_4H_4OH/MAG'

DP = DFTPartition(path)

DP.readVASPFolder(v=True)

DP.generateDFTVariants('Vacancy', [1], is_surface=True)
#DP.generateDFTVariants('KPOINTS', [[n,n,1] for n in range(1, 15)] ) 


path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/dataset/CoFeNiOOH_jingzhu/surf_CoFe_4H_4OH/MAG'
DP = DFTPartition(path)
DP.readVASPFolder(v=True)
DP.generateDFTVariants('NUPDOWN', [n for n in range(0, 10, 1)] )
DP.writePartition()

path = '/home/akaris/DocumeEENnts/code/Physics/VASP/v6.1/files/POSCAR/Cristals/NiOOH/*OH surface with Fe(HS)'
path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/NiOOH/*OH surface for pure NiOOH'
#DP = DFTPartition('/home/akaris/Documents/code/Physics/VASP/v6.1/files/bulk_optimization/Pt/parametros/ENCUT_optimization_252525_FCC')
DP = DFTPartition(path)
#DP.readVASPSubFolder(v=True)
DP.readVASPFolder(v=True)

#DP.generateDFTVariants('Vacancy', [1], is_surface=True)
#DP.generateDFTVariants('KPOINTS', [[n,n,1] for n in range(1, 15)] )    
DP.generateDFTVariants('ENCUT', [n for n in range(200, 1100, 45)] )

DP.writePartition()

DP.generateDFTVariants('ENCUT', [ E for E in range(400,700,30)] )
print( DP.summary() )

print( DP.containers[0].AtomPositionManager.summary() )
'''




try:
    from sage_lib.master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class KPointsManager(FileManager):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(name=name, file_location=file_location)
        self._num_kpoints = None
        self._coordinate_system = None
        self._kpoints = []
        self._mesh_type = None
        self._centering = None
        self._subdivisions = None
        self._shift = None
        self._points_per_line = None
        self._paths = []
        self._numberOfTetrahedra = None
        self._volumeWeight = None
        self._KPOINTmesh = None

    def readKPOINTS(self, file_location:str=None):
        file_location = file_location if type(file_location) == str else self.file_location

        lines = [n for n in self.read_file(file_location) ]

        self.comment = lines[0].strip()
        self.num_kpoints = int(float(lines[1].strip().split()[0]))

        mesh_type_flag = lines[2].strip().split()[0].lower()[0]    

        if self.num_kpoints == 0:
            if mesh_type_flag in ['g', 'm']:
                self.mesh_type = 'Regular'
                self.centering = 'Gamma' if mesh_type_flag == 'g' else 'Monkhorst-Pack'
                self.subdivisions = [int(float(n)) for n in lines[3].strip().split()[:3] ] #list(map(float, lines[3].strip().split()[:3]) )
                if len(lines) > 4:
                    try: self.shift = list(map(float, lines[4].strip().split()[:3] ))
                    except: pass

            elif mesh_type_flag in ['a']:
                self.mesh_type = 'Automatic'
                self.length = float(lines[3].strip())

            else:
                self.mesh_type = 'Generalized'
                self.coordinate_system = 'Cartesian' if mesh_type_flag in ['c', 'k'] else 'Fractional' if mesh_type_flag == 'r' else None
                self.KPOINTmesh = []
                for i, line in enumerate(lines[3:]):  
                    line_split = line.strip().split()
                    if len(line_split) > 2:
                        self.KPOINTmesh.append(line_split[:3]) 

        elif mesh_type_flag == 'l':
            self.mesh_type = 'Line Mode'
            self.points_per_line = self.num_kpoints
            self.coordinate_system = lines[3].strip()
            path = []
            for line in lines[4:]:
                if line.strip():  # Skip empty lines
                    coords, label = line.strip().rsplit(maxsplit=1)
                    kx, ky, kz = map(float, coords.split())
                    path.append({'kx': kx, 'ky': ky, 'kz': kz, 'label': label})
                else:
                    if path:
                        self.paths.append(path)
                        path = []
            if path:
                self.paths.append(path)  # Append remaining path if any

        else:
            self.mesh_type = 'Explicit'

            if mesh_type_flag == 'c' or mesh_type_flag == 'k':
                self.coordinate_system = 'Cartesian'
            else:        
                self.coordinate_system = 'Fractional'

            for i, line in enumerate(lines[3:]):  
                if line.strip().split()[0][0].lower() == 't': # Tetrahedron case
                    self.numberOfTetrahedra = lines[i+1].strip().split(' ')[0]
                    self.volumeWeight = lines[i+1].strip().split(' ')[1:]
                    break
                else:
                    kx, ky, kz, weight = map(float, line.split()[:4])
                    self.kpoints.append({'kx': kx, 'ky': ky, 'kz': kz, 'weight': weight})

    def exportAsKPOINTS(self, file_location:str=None):
        file_location = file_location if file_location is not None else self.file_location
        lines = []

        # Write the comment line
        lines.append(self.comment)

        # Write the number of k-points
        lines.append(str(self.num_kpoints))

        # Handle different mesh types
        if self.mesh_type == 'Regular':
            lines.append(self.centering)
            lines.append(" ".join(map(str, self.subdivisions)))
            if hasattr(self, 'shift') and self.shift:
                lines.append(" ".join(map(str, self.shift)))

        elif self.mesh_type == 'Automatic':
            lines.append('Auto')
            lines.append(str(self.length))

        elif self.mesh_type == 'Line Mode':
            lines.append('Line Mode')
            lines.append(self.coordinate_system)
            for path in self.paths:
                for point in path:
                    lines.append(f"{point[0]} {point[1]} {point[2]} ! {point[3]}")
                lines.append("")  # Empty line to separate different paths

        elif self.mesh_type == 'Explicit':
            lines.append(self.coordinate_system)
            for kpoint in self.kpoints:
                lines.append("{kx} {ky} {kz} {weight}".format(**kpoint))

        # Write to file
        with open(file_location, 'w') as f:
            for line in lines:
                f.write(line + "\n")

    def set_band_path(self, band_path, divisions:int=10):
        self.comment = 'KPOINTS for band structure'
        self.num_kpoints = divisions
        self.mesh_type = 'Line Mode'
        self.coordinate_system = 'Reciprocal'
        self.paths = [ [list(self.special_lattice_points[pp])+[pp], list(self.special_lattice_points[pi])+[pi]] for pp, pi in zip(band_path[::2],band_path[1::2] )]

    def summary(self, ) -> str:
        text_str = ''
        # Display parsed information based on the mesh type
        if self.mesh_type == 'Explicit':
            text_str +=f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Number of k-points: {self.num_kpoints}\n"
            text_str += f"Coordinate System: {self.coordinate_system}\n"
            text_str += f"K-points: {self.kpoints}\n"
            text_str += f"NumberOfTetrahedra: {self.numberOfTetrahedra}\n"
            text_str += f"VolumeWeight: {self.volumeWeight}\n"

        elif self.mesh_type == 'Line Mode':
            text_str += f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Points per Line: {self.points_per_line}\n"
            text_str += f"Coordinate System: {self.coordinate_system}\n"
            text_str += f"Paths:\n"
            for i, p in enumerate(self.paths):
                k0x,k0y,k0z,  k1x,k1y,k1z = p[0]['kx'],p[0]['ky'],p[0]['kz'],  p[1]['kx'],p[1]['ky'],p[1]['kz']  
                text_str +=  f' ({i})  {k0x:<4.2f},{k0y:<4.2f},{k0z:<4.2f} >> {k1x:<4.2f},{k1y:<4.2f},{k1z:<4.2f}\n'
 
        elif self.mesh_type == 'Automatic':
            text_str += f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Length: {self.length}\n"

        elif self.mesh_type == 'Generalized':
            text_str += f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Length: {self.coordinate_system}\n"
            text_str += f"Length: {self.KPOINTmesh}\n"
        else:
            text_str += f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Centering: {self.centering}\n"
            text_str += f"Subdivisions: {self.subdivisions}\n"
            if self.shift:
                text_str += f"Shift: {self.shift}\n"
        return text_str 
'''
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTband')
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTSgeneralized')
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTregular')
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTSexplicitTetrahedron')
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTSexplicit')
kpoints_manager.readKPOINTS()
print(kpoints_manager.summary())
kpoints_manager.exportKPOINTS('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTexportation')
'''
